/* $Id: copyarea.c,v 1.8 2006/01/28 16:46:10 cegger Exp $
***************************************************************************

   X server for LibGGI - GC (graphics context) acceleration

   Copyright (C) 1997      Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-2000 Marcus Sundberg	[marcus@ggi-project.org]
   Copyright (C) 2004-2006 Christoph Egger

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/


#include "gc.h"

static void doBitBlt(DrawablePtr pSrc, DrawablePtr pDst,
		GCPtr pGC, RegionPtr prgnDst, DDXPointPtr pptSrc);
static void doCopy(GCPtr pGC, int nbox,
		DDXPointPtr pptSrc, BoxPtr pbox,
		int xdir, int ydir);


/* Algorithm copied from cfb/cfbbitblt.c, cfbBitBlt() */

RegionPtr xggiCopyArea(
	     DrawablePtr pSrcDrawable,
	     DrawablePtr pDstDrawable,
	     GCPtr pGC,
	     int srcx, int srcy,
	     int width, int height,
	     int dstx, int dsty)
{
	RegionPtr prgnSrcClip = NULL;	/* may be a new region, or
					 * just a copy */
	Bool freeSrcClip = FALSE;

	RegionPtr       prgnExposed;
	RegionRec       rgnDst;
	DDXPointPtr pptSrc, ppt;
	register BoxPtr pbox;
	int             i;
	register int    dx;
	register int    dy;
	xRectangle      origSource;
	DDXPointRec     origDest;
	int             numRects;
	BoxRec          fastBox;

	int             fastClip = 0;	/* for fast clipping with pixmap
					 * source */
	int             fastExpose = 0;	/* for fast exposures with pixmap
					 * source */
	int canaccel;

	XGGI_GC_OP_PROLOGUE(pGC, pDstDrawable);

	origSource.x = srcx;
	origSource.y = srcy;
	origSource.width = width;
	origSource.height = height;
	origDest.x = dstx;
	origDest.y = dsty;

	canaccel = ((DRAWABLE_HAS_MASTER_VISUAL_BPP(pSrcDrawable, scr))
		&& (DRAWABLE_HAS_MASTER_VISUAL_BPP(pDstDrawable, scr))
		&& (pGC->alu == GXcopy)
		&& (pSrcDrawable->type != DRAWABLE_PIXMAP)
		&& (pDstDrawable->type != DRAWABLE_PIXMAP));

	if (!canaccel) {
		prgnExposed = pGC->ops->CopyArea(
					pSrcDrawable, pDstDrawable,
					pGC, srcx, srcy, width, height,
					dstx, dsty);
		XGGI_GC_OP_EPILOGUE(pGC, pDstDrawable);
		return prgnExposed;
	}	/* if */


	if ((pSrcDrawable != pDstDrawable) &&
	    pSrcDrawable->pScreen->SourceValidate) {
		(*pSrcDrawable->pScreen->SourceValidate)
			(pSrcDrawable, srcx, srcy, width, height);
	}

	srcx += pSrcDrawable->x;
	srcy += pSrcDrawable->y;

	/* clip the source */

	if (pSrcDrawable->type == DRAWABLE_PIXMAP) {
		if ((pSrcDrawable == pDstDrawable) &&
		    (pGC->clientClipType == CT_NONE)) {
			prgnSrcClip = pGC->pCompositeClip;
		} else {
			fastClip = 1;
		}
	} else {	/* Window */
		if (pGC->subWindowMode == IncludeInferiors) {
			/*
			 * XGGI DDX empties the border clip when the
			 * VT is inactive
			 */
			if (!((WindowPtr) pSrcDrawable)->parent &&
			    REGION_NOTEMPTY(pSrcDrawable->pScreen,
				 &((WindowPtr) pSrcDrawable)->borderClip)) {
				/*
				 * special case bitblt from root window in
				 * IncludeInferiors mode; just like from a pixmap
				 */
				fastClip = 1;
			} else if ((pSrcDrawable == pDstDrawable) &&
				   (pGC->clientClipType == CT_NONE)) {
				prgnSrcClip = pGC->pCompositeClip;
			} else {
				prgnSrcClip = NotClippedByChildren((WindowPtr) pSrcDrawable);
				freeSrcClip = TRUE;
			}
		} else {
			prgnSrcClip = &((WindowPtr) pSrcDrawable)->clipList;
		}
	}

	fastBox.x1 = srcx;
	fastBox.y1 = srcy;
	fastBox.x2 = srcx + width;
	fastBox.y2 = srcy + height;


	/* Don't create a source region if we are doing a fast clip */
	if (fastClip) {
		fastExpose = 1;
		/*
	         * clip the source; if regions extend beyond the source size,
	         * make sure exposure events get sent
	         */
		if (fastBox.x1 < pSrcDrawable->x) {
			fastBox.x1 = pSrcDrawable->x;
			fastExpose = 0;
		}
		if (fastBox.y1 < pSrcDrawable->y) {
			fastBox.y1 = pSrcDrawable->y;
			fastExpose = 0;
		}
		if (fastBox.x2 > pSrcDrawable->x + (int) pSrcDrawable->width) {
			fastBox.x2 = pSrcDrawable->x + (int) pSrcDrawable->width;
			fastExpose = 0;
		}
		if (fastBox.y2 > pSrcDrawable->y + (int) pSrcDrawable->height) {
			fastBox.y2 = pSrcDrawable->y + (int) pSrcDrawable->height;
			fastExpose = 0;
		}
	} else {
		REGION_INIT(pGC->pScreen, &rgnDst, &fastBox, 1);
		REGION_INTERSECT(pGC->pScreen, &rgnDst, &rgnDst, prgnSrcClip);
	}


	dstx += pDstDrawable->x;
	dsty += pDstDrawable->y;

	if (pDstDrawable->type == DRAWABLE_WINDOW) {
		if (!((WindowPtr) pDstDrawable)->realized) {
			if (!fastClip)
				REGION_UNINIT(pGC->pScreen, &rgnDst);
			if (freeSrcClip)
				REGION_DESTROY(pGC->pScreen, prgnSrcClip);

			XGGI_GC_OP_EPILOGUE(pGC, pDstDrawable);
			return NULL;
		}
	}
	dx = srcx - dstx;
	dy = srcy - dsty;

	/* Translate and clip the dst to the destination composite clip */
	if (fastClip) {
		RegionPtr       cclip;

		/* Translate the region directly */
		fastBox.x1 -= dx;
		fastBox.x2 -= dx;
		fastBox.y1 -= dy;
		fastBox.y2 -= dy;

		/*
		 * If the destination composite clip is one rectangle we can
		 * do the clip directly.  Otherwise we have to create a full
		 * blown region and call intersect
		 */

		/*
		 * XXX because CopyPlane uses this routine for 8-to-1 bit
		 * copies, this next line *must* also correctly fetch the
		 * composite clip from an mfb gc
		 */

		cclip = pGC->pCompositeClip;
		if (REGION_NUM_RECTS(cclip) == 1) {
			BoxPtr          pBox = REGION_RECTS(cclip);

			if (fastBox.x1 < pBox->x1) fastBox.x1 = pBox->x1;
			if (fastBox.x2 > pBox->x2) fastBox.x2 = pBox->x2;
			if (fastBox.y1 < pBox->y1) fastBox.y1 = pBox->y1;
			if (fastBox.y2 > pBox->y2) fastBox.y2 = pBox->y2;

			/* Check to see if the region is empty */
			if (fastBox.x1 >= fastBox.x2 || fastBox.y1 >= fastBox.y2) {
				REGION_NULL(pGC->pScreen, &rgnDst);
			} else {
				REGION_INIT(pGC->pScreen, &rgnDst, &fastBox, 1);
			}
		} else {
			/*
			 * We must turn off fastClip now, since we must
			 * create a full blown region.  It is intersected
			 * with the composite clip below.
			 */
			fastClip = 0;
			REGION_INIT(pGC->pScreen, &rgnDst, &fastBox, 1);
		}
	} else {
		REGION_TRANSLATE(pGC->pScreen, &rgnDst, -dx, -dy);
	}

	if (!fastClip) {
		REGION_INTERSECT(pGC->pScreen, &rgnDst, &rgnDst,
				 pGC->pCompositeClip);
	}


	/* Do bit blitting */
	numRects = REGION_NUM_RECTS(&rgnDst);

	if (numRects && width && height) {
		if (!(pptSrc = (DDXPointPtr)ALLOCATE_LOCAL(numRects *
						sizeof(DDXPointRec))))
		{
			REGION_UNINIT(pGC->pScreen, &rgnDst);
			if (freeSrcClip)
				REGION_DESTROY(pGC->pScreen, prgnSrcClip);
			XGGI_GC_OP_EPILOGUE(pGC, pDstDrawable);
			return NULL;
		}

		pbox = REGION_RECTS(&rgnDst);
		ppt = pptSrc;
		for (i = numRects; --i >= 0; pbox++, ppt++) {
			ppt->x = pbox->x1 + dx;
			ppt->y = pbox->y1 + dy;
		}

		doBitBlt(pSrcDrawable, pDstDrawable, pGC, &rgnDst, pptSrc);
		DEALLOCATE_LOCAL(pptSrc);
	}


	prgnExposed = NULL;
	if (pGC->fExpose) {
		/*
		 * Pixmap sources generate a NoExposed (we return NULL to do
		 * this)
		 */
		if (!fastExpose)
			prgnExposed =
				miHandleExposures(pSrcDrawable, pDstDrawable, pGC,
						  origSource.x, origSource.y,
						  (int) origSource.width,
						  (int) origSource.height,
					  origDest.x, origDest.y, GXcopy);
	}
	REGION_UNINIT(pGC->pScreen, &rgnDst);
	if (freeSrcClip)
		REGION_DESTROY(pGC->pScreen, prgnSrcClip);

	XGGI_GC_OP_EPILOGUE(pGC, pDstDrawable);
	return prgnExposed;

}	/* xggiCopyArea */



/* Copied from xorg's xaa/xaaBitBlt.c */
static void doBitBlt(DrawablePtr pSrc, DrawablePtr pDst,
		GCPtr pGC, RegionPtr prgnDst, DDXPointPtr pptSrc)
{
	int nbox, careful;
	BoxPtr pbox, pboxTmp, pboxNext, pboxBase, pboxNew1, pboxNew2;
	DDXPointPtr pptTmp, pptNew1, pptNew2;
	int xdir, ydir;

	/* XXX we have to err on the side of safety when both are windows,
	 * because we don't know if IncludeInferiors is being used.
	 */
	careful = ((pSrc == pDst) ||
		   ((pSrc->type == DRAWABLE_WINDOW) &&
		    (pDst->type == DRAWABLE_WINDOW)));

	pbox = REGION_RECTS(prgnDst);
	nbox = REGION_NUM_RECTS(prgnDst);

	pboxNew1 = NULL;
	pptNew1 = NULL;
	pboxNew2 = NULL;
	pptNew2 = NULL;
	if (careful && (pptSrc->y < pbox->y1)) {
		/* walk source bottom to top */
		ydir = -1;

		if (nbox > 1) {
			/* keep ordering in each hand, reverse order of bands */
			pboxNew1 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox);
			if (!pboxNew1) return;
			pptNew1 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
			if (!pptNew1) {
				DEALLOCATE_LOCAL(pboxNew1);
				return;
			}	/* if */

			pboxBase = pboxNext = pbox + nbox - 1;
			while (pboxBase >= pbox) {
				while ((pboxNext >= pbox) &&
					(pboxBase->y1 == pboxNext->y1))
				{
					pboxNext--;
				}	/* while */

				pboxTmp = pboxNext + 1;
				pptTmp = pptSrc + (pboxTmp - pbox);
				while (pboxTmp <= pboxBase) {
					*pboxNew1++ = *pboxTmp++;
					*pptNew1++ = *pptTmp++;
				}	/* while */
				pboxBase = pboxNext;
			}	/* while */
			pboxNew1 -= nbox;
			pbox = pboxNew1;
			pptNew1 -= nbox;
			pptSrc = pptNew1;
		}	/* if */
	} else {
		/* walk source top to bottom */
		ydir = 1;
	}	/* if */

	if (careful && (pptSrc->x < pbox->x1)) {
		/* walk source right tao left */
		xdir = -1;

		if (nbox > 1) {
			/* reverse order of rects in each band */
			pboxNew2 = (BoxPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
			pptNew2 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox);
			if (!pboxNew2 || !pptNew2) {
				if (pptNew2) DEALLOCATE_LOCAL(pptNew2);
				if (pboxNew2) DEALLOCATE_LOCAL(pboxNew2);
				if (pboxNew1) {
					DEALLOCATE_LOCAL(pptNew1);
					DEALLOCATE_LOCAL(pboxNew1);
				}	/* if */
				return;
			}	/* if */

			pboxBase = pboxNext = pbox;
			while (pboxBase < pbox + nbox) {
				while ((pboxNext < pbox + nbox) &&
					(pboxNext->y1 == pboxBase->y1))
				{
					pboxNext++;
				}	/* while */

				pboxTmp = pboxNext;
				pptTmp = pptSrc + (pboxTmp - pbox);
				while (pboxTmp != pboxBase) {
					*pboxNew2++ = *--pboxTmp;
					*pptNew2++ = *--pptTmp;
				}	/* while */
				pboxBase = pboxNext;
			}	/* while */
			pboxNew2 -= nbox;
			pbox = pboxNew2;
			pptNew2 -= nbox;
			pptSrc = pptNew2;
		}	/* if */
	} else {
		/* walk source left to right */
		xdir = 1;
	}	/* if */

	doCopy(pGC, nbox, pptSrc, pbox, xdir, ydir);

	if (pboxNew2) {
		DEALLOCATE_LOCAL(pptNew2);
		DEALLOCATE_LOCAL(pboxNew2);
	}
	if (pboxNew1) {
		DEALLOCATE_LOCAL(pptNew1);
		DEALLOCATE_LOCAL(pboxNew1);
	}

}	/* doBitBlt */


/* copied from xorg's xaa/xaaCpyArea.c */
static void doCopy(GCPtr pGC, int nbox,
		DDXPointPtr pptSrc, BoxPtr pbox,
		int xdir, int ydir)
{
	int scr;

	scr = xggiGetScreenIdx(pGC->pScreen);

	DBUF_PROLOGUE(scr);
	if ((xdir == ydir) && (xdir == 1)) {
		for (; nbox; pbox++, pptSrc++, nbox--) {
			ggiCopyBox(MASTER_VISUAL_VIS(scr),
				pptSrc->x, pptSrc->y,
				pbox->x2 - pbox->x1, pbox->y2 - pbox->y1,
				pbox->x1, pbox->y1);
		}	/* for */
		goto out;
	}	/* if */

	/*
	 * This is the case of a chip that only supports xdir = 1,
	 * with ydir = 1 or ydir = -1, but we have xdir = -1.
	 */
	for (; nbox; pbox++, pptSrc++, nbox--) {
		if (pptSrc->y != pbox->y1 || pptSrc->x >= pbox->x1) {
			/* No problem. Do a xdir = 1 blit instead. */
			ggiCopyBox(MASTER_VISUAL_VIS(scr),
				pptSrc->x, pptSrc->y,
				pbox->x2 - pbox->x1, pbox->y2 - pbox->y1,
				pbox->x1, pbox->y1);
		} else {
			/*
			 * This is the difficult case. Needs striping into
			 * non-overlapping horizontal chunks.
			 */
			int stripeWidth, w, fullStripes, extra, i;
			stripeWidth = 16;
			w = pbox->x2 - pbox->x1;
			if (pbox->x1 - pptSrc->x < stripeWidth)
				stripeWidth = pbox->x1 - pptSrc->x;
			fullStripes = w / stripeWidth;
			extra = w % stripeWidth;

			/* First, take care of the little bit on the far
			 * right
			 */
			if (extra) {
				ggiCopyBox(MASTER_VISUAL_VIS(scr),
					pptSrc->x + fullStripes * stripeWidth,
					pptSrc->y,
					extra, pbox->y2 - pbox->y1,
					pbox->x1 + fullStripes * stripeWidth,
					pbox->y1);
			}	/* if */

			/* Now, take care of the rest of the blit */
			for (i = fullStripes - 1; i >= 0; i--) {
				ggiCopyBox(MASTER_VISUAL_VIS(scr),
					pptSrc->x + i * stripeWidth,
					pptSrc->y,
					stripeWidth, pbox->y2 - pbox->y1,
					pbox->x1 + i * stripeWidth,
					pbox->y1);
			}	/* for */
		}	/* if */
	}	/* for */


out:
	DBUF_EPILOGUE(scr);
	return;
}	/* doCopy */
