/* $Id: getimage.c,v 1.2 2006/01/28 16:46:10 cegger Exp $
***************************************************************************

   X server for LibGGI - Graphics acceleration

   Copyright (C) 1997	   Michael Krause	[rawstyle@ms.demo.org]
   Copyright (C) 1998-1999 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************************************
*/

#include "screen.h"


/*
 * xggiGetImage - accelerated replacement for cfbGetImage
 *
 */
void xggiGetImage(DrawablePtr pDrawable,
			int sx, int sy, int w, int h,
			unsigned int format,
			unsigned long planeMask,
			char *pdstLine)
{
	int scr;
	PixmapPtr pPixmap;
	ScreenPtr pScreen;

	if (!cfbDrawableEnabled(pDrawable)) return;

	pScreen = pDrawable->pScreen;
	scr = xggiGetScreenIdx(pScreen);

	/* We can only accelerate GXcopy until
	 * libggiblt is available. *sigh*
	 */
	if ((DRAWABLE_HAS_MASTER_VISUAL_BPP(pDrawable, scr))
	  && (planeMask == GXcopy) && (format == ZPixmap))
	{
		BoxRec box;
		RegionRec rgnDst;

		pPixmap = GetScratchPixmapHeader(pDrawable->pScreen, w, h,
			pDrawable->depth, pDrawable->bitsPerPixel,
			PixmapBytePad(w,pDrawable->depth), (pointer)pdstLine);
		if (!pPixmap)
			return;

		box.x1 = 0;
		box.y1 = 0;
		box.x2 = w;
		box.y2 = h;
		REGION_INIT(pScreen, &rgnDst, &box, 1);
		DBUF_PROLOGUE(scr);
		ggiGetBox(MASTER_VISUAL_VIS(scr),
			sx + pDrawable->x, sy + pDrawable->y,
			w, h, (void *)pdstLine);
		DBUF_EPILOGUE(scr);

		REGION_UNINIT(pScreen, &rgnDst);
		FreeScratchPixmapHeader(pPixmap);
		return;
	}	/* if */

	/* Do XYPixmap and planeMask != GXcopy unaccelerated
	 */

	if (SCREEN_SCRIDX(scr) > 0) {
		SCREEN_SCRBACKUP(scr)->GetImage(pDrawable, sx, sy, w, h,
					format, planeMask, pdstLine);
	} else {
		SCREEN_SCR(scr)->GetImage(pDrawable, sx, sy, w, h,
					format, planeMask, pdstLine);
	}	/* if */

}	/* xggiGetImage */
